
import asyncio
import enum
import functools
import typing
import aiohttp

from loguru import logger

from .event import ToolEvent

from AgentService.enums.connector import ToolResponse
from AgentService.types.response import AgentResponse, AgentResponseType
from AgentService.dtypes.message import SendMessageRequest, Tool, ToolAnswer


class AgentConnector:
    def __init__(
        self,
        endpoint: str,
        key: str = None
    ):

        self.log = logger.bind(classname=self.__class__.__name__)

        self.endpoint = endpoint[:-1] if endpoint.endswith("/") else endpoint
        self.key = key

        self.callbacks = {}

    def bind_tool_output(self, tool_name: str, function: typing.Callable):
        self.callbacks.update({tool_name: functools.partial(self.answer_tool, function=function)})
        self.log.debug(f"Binded callback -> {tool_name} -> {function}")

    async def add_to_storage(self, key: str, data: list[dict] | dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/add",
                json={
                    "key": key,
                    "data": data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def remove_from_storage(self, key: str, data: list[dict] | dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/remove",
                json={
                    "key": key,
                    "data": data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def update_storage(self, key: str, data: dict, new_data: dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/update",
                json={
                    "key": key,
                    "data": data,
                    "new_data": new_data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def get_from_storage(self, key: str) -> list[dict]:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=f"{self.endpoint}/storage",
                json={
                    "key": key
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def answer_tool(self, event: ToolEvent, function: typing.Callable) -> ToolAnswer:
        try:
            response = await function(data=event.tool.arguments, chat_id=event.chat_id)

            if response in [None, ""]:
                response = "function returned nothing"

            elif isinstance(response, enum.Enum):
                response = response.value

            elif not isinstance(response, str):
                response = str(response)

            return ToolAnswer(
                id=event.tool.id,
                name=event.tool.name,
                text=response
            )

        except Exception as err:
            self.log.exception(err)

            return ToolAnswer(
                id=event.tool.id,
                name=event.tool.name,
                text="function returned nothing"
            )

    async def dispatch_event(self, event: ToolEvent):
        callback = self.callbacks.get(event.tool.name)

        if not callback:
            self.log.warning(f"Can't find callback function \"{event.tool.name}\" in {self.callbacks}")

            return ToolAnswer(
                id=event.tool.id,
                name=event.tool.name,
                text="function not binded"
            )

        while True:
            response: ToolAnswer = await callback(event=event)
            self.log.debug(f"Dispatching event -> {event} -> {callback} -> {response}")

            try:
                ToolResponse[response.text]

            except KeyError:
                break

            await asyncio.sleep(5)

        return response

    async def handle_tools(self, tools: list[Tool], chat_id) -> list[ToolAnswer]:
        tool_answers = [
            await self.dispatch_event(ToolEvent(tool_name=tool.name, tool=tool, chat_id=chat_id))
            for tool in tools
        ]

        return tool_answers

    async def request(self, request: SendMessageRequest) -> AgentResponse:
        self.log.debug(f"Sending request -> {request.to_dict()}")

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/chat",
                json=request.to_dict(),
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")

                response = await resp.json()
                agent_response = AgentResponse(**response["data"])

                return agent_response

    async def send(self, chat_id: str, text: str = None, context: dict = {}, tool_answers: list[ToolAnswer] = []) -> AgentResponse:
        agent_response = await self.request(SendMessageRequest(chat_id=chat_id, text=text, context=context, tool_answers=tool_answers))

        if agent_response.type == AgentResponseType.answer:
            return agent_response

        tool_answers = await self.handle_tools(agent_response.content, chat_id)

        return await self.send(chat_id, tool_answers=tool_answers)




from typing import Dict, List
from pydantic import BaseModel


class GetStorage(BaseModel):
    key: str


class AddStorage(BaseModel):
    key: str
    data: List[Dict] | Dict


class RemoveStorage(BaseModel):
    key: str
    data: List[Dict] | Dict


class UpdateStorage(BaseModel):
    key: str
    data: Dict
    new_data: Dict


class SearchStorage(BaseModel):
    query: str




from AgentService.dtypes.db import DatabaseItem


class Storage(DatabaseItem):
    def __init__(
        self,
        id: str,
        key: str
    ):

        self.id = id
        self.key = key

        self.fields = ["id", "key"]


class StorageItem(DatabaseItem):
    def __init__(
        self,
        id: str,
        key: str,
        data: dict
    ):

        self.id = id
        self.key = key
        self.data = data

        self.fields = ["id", "key", "data"]




import fastapi
import io
import json

from .models import GetStorage, AddStorage, RemoveStorage, UpdateStorage, SearchStorage

from AgentService.config import Config
from AgentService.db import Db
from AgentService.dtypes.db import method as dmth
from AgentService.dtypes.storage import StorageItem, Storage
from AgentService.dtypes.response import OkResponse, ErrResponse
from AgentService.utils.convert import string_to_uuid


storage_router = fastapi.APIRouter(prefix="/storage")
db = Db()


async def get_packed_storage(key: str) -> io.BytesIO:
    storage_items: list[StorageItem] = await db.ex(dmth.GetMany(StorageItem, key=key))
    data = [
        {
            "text": item.data,
            "metadata": {"id": item.id}
        }
        for item in storage_items
    ]

    buffer = io.BytesIO()
    buffer.write((json.dumps(data, ensure_ascii=False)).encode("utf-8"))
    buffer.seek(0)

    return buffer


@storage_router.get("")
async def get_storage(request: GetStorage):
    key = request.key

    items: list[StorageItem] = await db.ex(dmth.GetMany(StorageItem, key=key))

    return OkResponse(
        data=items
    ).to_dict()


@storage_router.post("/add")
async def add_to_storage(request: AddStorage):
    key = request.key
    data = request.data

    agent = Config().agent

    if not isinstance(data, list):
        data = [data]

    storage: Storage = await db.ex(dmth.GetOne(Storage, key=key))
    if not storage:
        storage = await agent.create_storage(key=key)
        await db.ex(dmth.AddOne(Storage, storage))

    storage_items = [
        StorageItem(
            id=string_to_uuid(key+str(row)),
            key=key,
            data=row
        )
        for row in data
    ]
    await db.ex(dmth.AddMany(StorageItem, storage_items))

    buffer = await get_packed_storage(key)
    await agent.update_storage(buffer, storage.id)

    return OkResponse().to_dict()


@storage_router.post("/remove")
async def remove_from_storage(request: RemoveStorage):
    key = request.key
    data = request.data

    agent = Config().agent

    if not isinstance(data, list):
        data = [data]

    for item in data:
        item_id = string_to_uuid(key+str(item))

        storage_item: StorageItem = await db.ex(dmth.GetOne(StorageItem, key=key, id=item_id))
        if not storage_item:
            return ErrResponse(
                description=f"No such storage item -> {key} -> {item}@{item_id}"
            ).to_dict()

        await db.ex(dmth.RemoveOne(StorageItem, storage_item))

    storage: Storage = await db.ex(dmth.GetOne(Storage, key=key))
    buffer = await get_packed_storage(key)
    await agent.update_storage(buffer, storage.id)

    return OkResponse().to_dict()


@storage_router.post("/update")
async def update_storage(request: UpdateStorage):
    key = request.key
    data = request.data
    new_data = request.new_data

    agent = Config().agent

    item_id = string_to_uuid(key+str(data))

    storage_item: StorageItem = await db.ex(dmth.GetOne(StorageItem, key=key, id=item_id))
    if not storage_item:
        return ErrResponse(
            description=f"No such storage item -> {key} -> {data}:{item_id}"
        ).to_dict()

    await db.ex(dmth.RemoveOne(StorageItem, storage_item))

    storage_item.id = string_to_uuid(str(new_data))
    storage_item.data = new_data
    await db.ex(dmth.AddOne(StorageItem, storage_item))

    storage: Storage = await db.ex(dmth.GetOne(Storage, key=key))
    buffer = await get_packed_storage(key)
    await agent.update_storage(buffer, storage.id)

    return OkResponse().to_dict()
